import csv
import itertools
import os
import shutil
import tempfile
from datetime import datetime
import pytest
import lucene
from lupyne import engine

dirname = os.path.join(os.path.dirname(__file__), 'fixtures')


def pytest_report_header(config):
    return 'PyLucene ' + lucene.VERSION


def pytest_configure(config):
    assert lucene.initVM()


@pytest.fixture
def tempdir():
    tempdir = tempfile.mkdtemp(dir=dirname)
    yield tempdir
    shutil.rmtree(tempdir)


def fixture(gen):
    return pytest.fixture(lambda: gen())


@pytest.fixture
def fields():
    return [
        engine.Field.Text('text', storeTermVectors=True, storeTermVectorPositions=True, storeTermVectorOffsets=True),
        engine.Field.String('article', stored=True),
        engine.Field.String('amendment', stored=True),
        engine.Field.String('date', stored=True, docValuesType='sorted'),
        engine.Field('year', docValuesType='numeric'),
    ]


@fixture
def constitution():
    lines = open(os.path.join(dirname, 'constitution.txt'))
    items = itertools.groupby(lines, lambda l: l.startswith('Article ') or l.startswith('Amendment '))
    for _, (header,) in items:
        _, lines = next(items)
        header, num = header.rstrip('.\n').split(None, 1)
        fields = {header.lower(): num, 'text': ''.join(lines)}
        if header == 'Amendment':
            num, date = num.split()
            date = datetime.strptime(date, '%m/%d/%Y').date()
            fields.update({header.lower(): num, 'date': str(date), 'year': date.year})
        yield fields


@fixture
def zipcodes():
    lines = open(os.path.join(dirname, 'zipcodes.txt'))
    for zipcode, latitude, longitude, state, city, county in csv.reader(lines):
        yield {'zipcode': zipcode, 'latitude': float(latitude), 'longitude': float(longitude), 'city': city.title(), 'county': county.title(), 'state': state}



import calendar
import http
import json
import math
import operator
import os
import signal
import subprocess
import sys
import time
from email.utils import parsedate
import cherrypy
import clients
import portend
import pytest
from lupyne import engine, server


class Servers(dict):
    module = 'lupyne.server'
    ports = 8080, 8081, 8082
    config = {'server.socket_timeout': 2, 'server.shutdown_timeout': 1}

    def start(self, port, *args, **config):
        config.update(self.config)
        config['server.socket_port'] = port
        portend.free('localhost', port)
        server = self[port] = subprocess.Popen((sys.executable, '-m', self.module, '-c', json.dumps(config)) + args)
        portend.occupied('localhost', port)
        server.started = time.time()
        assert not server.poll()
        return clients.Resource('http://localhost:{}'.format(port))

    def stop(self, port):
        server = self.pop(port)
        time.sleep(max(0, server.started + 0.2 - time.time()))  # let server finish starting for clean shutdown
        server.terminate()
        assert server.wait() == 0


@pytest.fixture
def servers(request):
    servers = Servers()
    servers.config['log.screen'] = request.config.option.verbose > 0
    yield servers
    for port in list(servers):
        servers.stop(port)


@pytest.fixture
def resource(tempdir, servers, fields, constitution):
    resource = servers.start(servers.ports[0], tempdir)
    for field in fields:
        assert resource.client.put('fields/' + field.name, field.settings).status_code == http.client.CREATED
    assert resource.post('docs', list(constitution)) is None
    assert resource.post('update', {'spellcheckers': True, 'merge': 1})
    return resource


def test_docs(resource):
    ((directory, size),) = resource().items()
    assert 'Directory@' in directory and size == 35
    fields = resource.get('/fields')
    assert sorted(fields) == ['amendment', 'article', 'date', 'text', 'year']
    for field in ('amendment', 'article', 'date'):
        assert not resource.fields(field)['tokenized']
    assert resource.fields('text')['storeTermVectors']
    assert resource.fields('year')['docValuesType'] == 'NUMERIC'
    assert resource.client.put('fields/' + field, {'stored': True}).status_code == http.client.OK
    assert resource.docs('8', **{'fields.docvalues': 'date'}) == {'amendment': '1', 'date': '1791-12-15'}
    assert resource.docs('8', fields='', **{'fields.docvalues': 'year:int'}) == {'year': 1791}
    assert resource.client.get('docs/8', params={'fields.docvalues': 'year'}).status_code == http.client.BAD_REQUEST
    assert resource.client.get('docs/8', params={'fields.docvalues': 'year:invalid'}).status_code == http.client.BAD_REQUEST
    doc = resource.docs('0', **{'fields.vector': 'text,missing'})
    assert doc['missing'] == [] and doc['text'].index('states') < doc['text'].index('united')
    doc = resource.docs('0', **{'fields.vector.counts': 'text'})
    resource.client.patch('docs/amendment/1', {'amendment': '1'}).status_code == http.client.CONFLICT
    assert not resource.patch('docs/amendment/1')
    assert {term for term, count in doc['text'].items() if count > 1} >= {'establish', 'states', 'united'}
    assert resource.client.put('docs/article/0', {'article': '-1'}).status_code == http.client.BAD_REQUEST
    assert resource.delete('docs/article/0') is None


def test_terms(resource):
    assert resource.terms() == ['amendment', 'article', 'date', 'text', 'year']
    articles = resource.terms('article')
    articles.remove('Preamble')
    assert sorted(map(int, articles)) == list(range(1, 8))
    assert sorted(map(int, resource.terms('amendment'))) == list(range(1, 28))
    assert resource.terms('text/:0') == []
    assert resource.terms('text/right:right~') == resource.terms('text/right*') == ['right', 'rights']
    assert resource.terms('text/writ*') == ['writ', 'writing', 'writings', 'writs', 'written']
    assert resource.terms('text/writ*?count=10') == ['writs', 'writ', 'writing', 'writings', 'written']
    assert resource.terms('text/writ*?count=3') == ['writs', 'writ', 'writing']
    assert resource.terms('text/right~1') == ['eight', 'right', 'rights']
    assert resource.terms('text/right~') == ['eight', 'high', 'right', 'rights']
    assert resource.terms('text/right~?count=3') == []
    assert resource.terms('text/write~?count=3') == ['writs', 'writ', 'written']
    docs = resource.terms('text/people/docs')
    assert resource.terms('text/people') == len(docs) == 8
    counts = dict(resource.terms('text/people/docs/counts'))
    assert sorted(counts) == docs and all(counts.values()) and sum(counts.values()) > len(counts)
    positions = dict(resource.terms('text/people/docs/positions'))
    assert sorted(positions) == docs and list(map(len, positions.values())) == list(counts.values())
    assert resource.client.get('terms/text/people/missing').status_code == http.client.NOT_FOUND
    assert resource.client.get('terms/text/people/docs/missing').status_code == http.client.NOT_FOUND


def test_search(resource):
    (doc,) = resource.search(q='amendment:1', fields='', **{'fields.docvalues': 'date,year:int'})['docs']
    assert doc['date'] == '1791-12-15' and doc['year'] == 1791
    result = resource.search(**{'q.field': 'text', 'q': 'write "hello world"', 'q.spellcheck': 'true'})
    assert result['query'] == 'text:writs text:"held would"'
    assert result['count'] == len(result['docs']) == resource.get('/terms/text/writs') == 2
    assert resource.search(q='Preamble', **{'q.field': 'article'})['count'] == 0
    result = resource.search(q='Preamble', **{'q.field': 'article', 'q.type': 'prefix'})
    assert result['count'] == 1 and result['query'] == 'article:Preamble*'
    result = resource.search(q='text:"We the People"', **{'q.phraseSlop': 3})
    assert result['count'] == 1
    assert result['query'].startswith('text:"we ') and result['query'].endswith(' people"~3')
    (doc,) = result['docs']
    assert sorted(doc) == ['__id__', '__score__', 'article']
    assert doc['article'] == 'Preamble' and doc['__id__'] >= 0 and 0 < doc['__score__']
    result = resource.search(q='text:people')
    docs = result['docs']
    assert sorted(docs, key=operator.itemgetter('__score__'), reverse=True) == docs
    assert len(docs) == result['count'] == 8
    result = resource.search(q='text:people', count=5)
    assert docs[:5] == result['docs'] and result['count'] == len(docs)
    result = resource.search(q='text:people', count=5, sort='-year:int')
    assert all(math.isnan(doc['__score__']) for doc in result['docs'])
    assert result['docs'][0]['__sortkeys__'] == [1913] and result['docs'][-1]['__sortkeys__'] == [1791]
    result = resource.search(q='text:people', sort='-year:int')
    assert result['docs'][0]['__sortkeys__'] == [1913] and result['docs'][-1]['__sortkeys__'] == [0]
    result = resource.search(q='text:people', count=5, sort='-year:int', **{'sort.scores': ''})
    result = resource.search(q='text:people', count=1, sort='-year:int', **{'sort.scores': 'max'})
    result = resource.search(q='text:people', count=5, sort='-date,year:int')
    assert result['docs'][0]['__sortkeys__'] == ['1913-04-08', 1913] and result['docs'][-1]['__sortkeys__'] == ['1791-12-15', 1791]
    result = resource.search(q='text:people', start=2, count=2, facets='date')
    assert result['count'] == 8 and result['facets']['date'] == {'1791-12-15': 5, '1913-04-08': 1}
    result = resource.search(q='text:president', facets='date')
    assert len(result['facets']['date']) == sum(result['facets']['date'].values()) == 7
    result = resource.search(q='text:freedom')
    assert result['count'] == 1
    (doc,) = result['docs']
    assert doc['amendment'] == '1'
    assert resource.delete('search', params={'q': 'text:freedom'}) is None
    assert resource.post('update') == 34
    assert resource.search(q='text:freedom')['docs'] == []


def test_highlights(resource):
    (doc,) = resource.search(q='amendment:1', hl='amendment', fields='article')['docs']
    assert doc['__highlights__'] == {'amendment': '<b>1</b>'}
    result = resource.search(q='text:1', hl='amendment,article')
    highlights = [doc['__highlights__'] for doc in result['docs']]
    assert all(highlight['amendment'] or highlight['article'] for highlight in highlights)
    result = resource.search(mlt=0)
    assert result['count'] >= 25 and set(result['query'].split()) >= {'text:united', 'text:states'}
    result = resource.search(q='amendment:2', mlt=0, **{'mlt.fields': 'text', 'mlt.minTermFreq': 1, 'mlt.minWordLen': 6})
    assert result['count'] == 11 and set(result['query'].split()) == {'text:necessary', 'text:people'}
    assert [doc['amendment'] for doc in result['docs'][:3]] == ['2', '9', '10']
    result = resource.search(q='text:people', count=1, timeout=-1)
    assert result == {'query': 'text:people', 'count': 0, 'docs': []}
    result = resource.search(q='text:people', timeout=0.01)
    assert result['count'] in (0, 8)
    result = resource.search(q='+text:right +text:people')
    assert result['count'] == 4
    assert resource.search(q='hello', **{'q.field': 'body.title^2.0'})['query'] == '(body.title:hello)^2.0'


def test_groups(resource):
    result = resource.search(q='text:right', group='date', count=2, **{'group.count': 2})
    assert 'docs' not in result and len(result['groups']) == 2
    assert sum(map(operator.itemgetter('count'), result['groups'])) < result['count'] == 13
    assert all(min(group['count'], 2) >= len(group['docs']) for group in result['groups'])
    assert all(doc.get('date') == group['value'] for group in result['groups'] for doc in group['docs'])
    group = result['groups'][0]
    assert group['value'] == '1791-12-15'
    assert sorted(group) == ['count', 'docs', 'value'] and group['count'] == 5
    assert len(group['docs']) == 2 and group['docs'][0]['amendment'] == '2'
    assert len(result['groups'][1]['docs']) == 1 and all(group['docs'] == [] for group in result['groups'][2:])
    result = resource.search(q='text:right', group='year:int')
    groups = {group['value']: group['count'] for group in result['groups']}
    assert len(groups) == result['count'] == 9 < sum(groups.values())
    group = result['groups'][0]
    assert group['value'] == 1791 and group['count'] == 5
    assert resource.client.get('search', params={'q': 'text:right', 'group': 'date:int'}).status_code == http.client.BAD_REQUEST


def test_facets(tempdir, servers, zipcodes):
    writer = engine.IndexWriter(tempdir)
    writer.commit()
    resource = servers.start(servers.ports[0], '-r', tempdir)
    writer.set('zipcode', dimensions=1, stored=True)
    writer.fields['location'] = engine.NestedField('county.city', docValuesType='sorted')
    for doc in zipcodes:
        if doc['state'] == 'CA':
            writer.add(zipcode=int(doc['zipcode']), location='{}.{}'.format(doc['county'], doc['city']))
    writer.commit()
    assert resource.post('update') == resource().popitem()[1] == len(writer)
    result = resource.search(count=0, facets='county', **{'count.min': 10000})
    facets = result['facets']['county']
    assert result['count'] == sum(facets.values()) and 'Los Angeles' in facets
    result = resource.search(q='Los Angeles', count=0, facets='county.city', **{'q.type': 'term', 'q.field': 'county'})
    facets = result['facets']['county.city']
    assert result['count'] == sum(facets.values()) and all(location.startswith('Los Angeles.') for location in facets)
    result = resource.search(count=0, facets='county', **{'facets.count': 3})
    assert sorted(result['facets']['county']) == ['Los Angeles', 'Orange', 'San Diego']
    result = resource.search(count=0, facets='county', **{'facets.min': 140})
    assert sorted(result['facets']['county']) == ['Los Angeles', 'Orange', 'San Diego']
    result = resource.search(q='Los Angeles', group='county.city', **{'group.count': 2, 'q.field': 'county', 'q.type': 'prefix'})
    assert all(group['value'].startswith('Los Angeles') for group in result['groups'])
    assert sum(map(operator.itemgetter('count'), result['groups'])) == sum(facets.values()) == result['count']


def test_queries(servers):
    resource = servers.start(servers.ports[0], **{'tools.validate.etag': False})
    resource.headers['content-length'] = '0'
    assert resource.queries() == []
    resource.client.get('queries/default').status_code == http.client.NOT_FOUND
    resource.client.put('queries/default/alpha', '*:*').status_code == http.client.CREATED
    assert resource.put('queries/default/alpha', 'name:alpha') == 'name:alpha'
    assert resource.put('queries/default/bravo', 'name:bravo') == 'name:bravo'
    assert resource.queries() == ['default']
    assert resource.queries('default') == {'alpha': 0.0, 'bravo': 0.0}
    assert resource.queries('default/alpha') == 'name:alpha'
    resource.client.get('queries/charlie').status_code == http.client.NOT_FOUND
    queries = resource.post('queries/default', {'name': 'alpha'})
    assert queries['alpha'] > 0.0 and queries['bravo'] == 0.0
    queries = resource.post('queries/default', {'name': 'alpha bravo alpha'})
    assert queries['alpha'] > queries['bravo'] > 0.0
    assert resource.delete('queries/default/alpha') == 'name:alpha'
    assert resource.delete('queries/default/alpha') is None
    assert resource.queries('default') == {'bravo': 0.0}


def test_realtime(tempdir, servers):
    for args in [('-r',), ('--real-time', 'index0', 'index1'), ('-r', '--real-time', 'index')]:
        assert subprocess.call((sys.executable, '-m', 'lupyne.server') + args, stderr=subprocess.PIPE)
    root = server.WebIndexer(tempdir)
    root.indexer.add()
    assert root.update() == 1
    del root
    resource = servers.start(servers.ports[0], '--real-time', **{'tools.validate.expires': 0})
    client = resource.client
    response = client.get('docs')
    version, modified, expires = map(response.headers.get, ('etag', 'last-modified', 'expires'))
    assert response.ok and modified is None and response.json() == []
    assert client.post('docs', [{}]).status_code == http.client.OK
    assert resource.docs() == [0]
    assert client.delete('search').status_code == http.client.OK
    assert resource.docs() == []
    time.sleep(max(0, calendar.timegm(parsedate(expires)) + 1 - time.time()))
    assert client.post(json=[tempdir]).status_code == http.client.OK
    response = client.get('docs')
    assert response.ok and response.json() == [0] and expires != response.headers['expires']
    assert client.post('update').ok
    response = client.get('docs')
    assert response.ok and version != response.headers['etag']


def test_start(tempdir, servers):
    port = servers.ports[0]
    pidfile = os.path.join(tempdir, 'pid')
    servers.start(port, '-dp', pidfile)
    time.sleep(1)
    os.kill(int(open(pidfile).read()), signal.SIGTERM)
    del servers[port]
    assert subprocess.call((sys.executable, '-m', 'lupyne.server', '-c', __file__), stderr=subprocess.PIPE)
    assert cherrypy.tree.mount(None)
    server.init(vmargs=None)
    with pytest.raises(AttributeError):
        server.start(config=True)


def test_config(tempdir, servers):
    engine.IndexWriter(tempdir).close()
    config = {'tools.validate.last_modified': True, 'tools.validate.expires': 0, 'tools.validate.max_age': 0}
    client = servers.start(servers.ports[0], tempdir, tempdir, '--autoreload=0.1', **config).client
    response = client.get()
    assert response.ok
    ((directory, size),) = response.json().items()
    assert 'Directory@' in directory and size == 0
    assert int(response.headers['age']) >= 0
    assert response.headers['cache-control'] == 'max-age=0'
    assert float(response.headers['x-response-time']) > 0.0
    dates = [parsedate(response.headers[header]) for header in ('last-modified', 'expires', 'date')]
    assert all(dates) and sorted(dates) == dates
    w, verion, _ = response.headers['etag'].split('"')
    assert w == 'W/' and verion.isdigit()


def test_example(request, servers):
    """Custom server example (only run explicitly)."""
    if request.config.option.verbose < 0:
        pytest.skip("requires verbose output")
    servers.module = 'tests.server'
    resource = servers.start(servers.ports[0])
    result = resource.search(q='date:17*', group='date')
    assert {group['value']: group['count'] for group in result['groups']} == {'1795-02-07': 1, '1791-12-15': 10}
    result = resource.search(count=0, facets='year')
    facets = result['facets']['year']
    assert not result['docs'] and facets['1791'] == 10 and sum(facets.values()) == result['count']
    result = resource.search(q='text:right', facets='year')
    facets = result['facets']['year']
    assert len(result['docs']) == result['count'] == sum(facets.values())


def test_replication(tempdir, servers):
    directory = os.path.join(tempdir, 'backup')
    primary = servers.start(servers.ports[0], tempdir)
    sync, update = '--autosync=' + primary.url, '--autoupdate=1'
    secondary = servers.start(servers.ports[1], '-r', directory, sync, update)
    resource = servers.start(servers.ports[2], '-r', directory)
    for args in [('-r', tempdir), (update, tempdir), (update, tempdir, tempdir)]:
        assert subprocess.call((sys.executable, '-m', 'lupyne.server', sync) + args, stderr=subprocess.PIPE)
    primary.post('docs', [{}])
    assert primary.post('update') == 1
    assert primary.client.put('update/0').status_code == http.client.NOT_FOUND
    response = resource.client.post(json={'url': primary.url})
    assert response.status_code == http.client.ACCEPTED and sum(response.json().values()) == 0
    assert resource.post('update') == 1
    assert resource.post(json={'url': primary.url})
    assert resource.post('update') == 1
    primary.post('docs', [{}])
    assert primary.post('update') == 2
    time.sleep(1.1)
    assert sum(secondary().values()) == 2
    servers.stop(servers.ports[-1])
    root = server.WebSearcher(directory, urls=(primary.url, secondary.url))
    app = server.mount(root)
    root.fields = {}
    assert root.update() == 2
    assert len(root.urls) == 2
    servers.stop(servers.ports[0])
    assert secondary.docs()
    assert secondary.client.post('docs', []).status_code == http.client.METHOD_NOT_ALLOWED
    assert secondary.terms() == []
    assert root.update() == 2
    assert len(root.urls) == 1
    servers.stop(servers.ports[1])
    assert root.update() == 2
    assert len(root.urls) == 0 and isinstance(app.root, server.WebIndexer)
    app.root.close()
    root = server.WebSearcher(directory)
    app = server.mount(root, autoupdate=0.1)
    root.fields, root.autoupdate = {}, 0.1
    cherrypy.config['log.screen'] = servers.config['log.screen']
    cherrypy.engine.state = cherrypy.engine.states.STARTED
    root.monitor.start()  # simulate engine starting
    time.sleep(0.2)
    app.root.indexer.add()
    time.sleep(0.2)
    assert len(app.root.indexer) == len(root.searcher) + 1
    app.root.monitor.unsubscribe()
    del app.root



"""
Example server.

Fields settings are assigned directly to the root.
Indexing is done here just to populate the example.

Registered queries demonstrate custom facets.

Example queries:
 * http://localhost:8080/search?q=date:17*&group=date
 * http://localhost:8080/search?count=0&facets=year
 * http://localhost:8080/search?q=text:right&count=3&facets=year
"""

import lucene
from lupyne import engine, server
from tests import conftest

Q = engine.Query


if __name__ == '__main__':
    assert lucene.getVMEnv() or lucene.initVM(vmargs='-Xrs')
    root = server.WebIndexer()
    # assign field settings
    root.indexer.set('amendment', engine.Field.String, stored=True)
    root.indexer.set('date', engine.Field.String, stored=True, docValuesType='sorted')
    root.indexer.set('text', engine.Field.Text)
    # populate index
    for doc in conftest.constitution.__wrapped__():
        if 'amendment' in doc:
            root.indexer.add(amendment=doc['amendment'], date=doc['date'], text=doc['text'])
    root.update()
    # assign custom facet queries based on year
    years = {date.split('-')[0] for date in root.searcher.terms('date')}
    root.query_map['year'] = {year: Q.prefix('date', year) for year in years}
    # start server
    server.start(root)



"""
Restful json `CherryPy <http://cherrypy.org/>`_ server.

The server script mounts a `WebSearcher`_ (read_only) or `WebIndexer`_ root.
Standard `CherryPy configuration <http://docs.cherrypy.org/en/latest/config.html>`_ applies,
and the provided `custom tools <#tools>`_ are also configurable.
All request and response bodies are `application/json values <http://tools.ietf.org/html/rfc4627.html#section-2.1>`_.

WebSearcher exposes resources for an IndexSearcher.
In addition to search requests, it provides access to term and document information in the index.

 * :meth:`/ <WebSearcher.index>`
 * :meth:`/search <WebSearcher.search>`
 * :meth:`/docs <WebSearcher.docs>`
 * :meth:`/terms <WebSearcher.terms>`
 * :meth:`/update <WebSearcher.update>`
 * :meth:`/queries <WebSearcher.queries>`

WebIndexer extends WebSearcher, exposing additional resources and methods for an Indexer.
Single documents may be added, deleted, or replaced by a unique indexed field.
Multiples documents may also be added or deleted by query at once.
By default changes are not visible until the update resource is called to commit a new index version.
If a near real-time Indexer is used, then changes are instantly searchable.
In such cases a commit still hasn't occurred, and the index based :meth:`last-modified header <validate>` shouldn't be used for caching.

 * :meth:`/ <WebIndexer.index>`
 * :meth:`/search <WebIndexer.search>`
 * :meth:`/docs <WebIndexer.docs>`
 * :meth:`/fields <WebIndexer.fields>`
 * :meth:`/update <WebIndexer.update>`

Custom servers should create and mount WebSearchers and WebIndexers as needed.
:meth:`Caches <WebSearcher.update>` and :meth:`field settings <WebIndexer.fields>` can then be applied directly before `starting <#start>`_ the server.
WebSearchers and WebIndexers can of course also be subclassed for custom interfaces.

CherryPy and Lucene VM integration issues:

 * Monitors (such as autoreload) are not compatible with the VM unless threads are attached.
 * WorkerThreads must be also attached to the VM.
 * VM initialization must occur after daemonizing.
 * Recommended that the VM ignores keyboard interrupts (-Xrs) for clean server shutdown.
"""

import argparse
import collections
import contextlib
import heapq
import http
import itertools
import os
import re
import time
import warnings
import lucene
import cherrypy
import clients
from requests.compat import json
from lupyne import engine

cherrypy.tools.params._priority = 15  # fix for json_out compatibility


def HTTPError(exception, status=http.client.BAD_REQUEST):
    return cherrypy.HTTPError.handle(exception, status)


@cherrypy.tools.register('before_request_body')
def json_in(process_body=None, **kwargs):
    """Handle request bodies in json format.

    :param content_type: request media type
    :param process_body: optional function to process body into request.params
    """
    request = cherrypy.serving.request

    def processor(entity):
        cherrypy.lib.jsontools.json_processor(entity)
        if process_body is not None:
            with HTTPError(TypeError):
                request.params.update(process_body(request.json))

    cherrypy.lib.jsontools.json_in(force='content-type' in request.headers, processor=processor, **kwargs)


@cherrypy.tools.register('before_handler')
def json_out(content_type='application/json', **kwargs):
    """Handle responses in json format.

    :param content_type: response content-type header
    """

    def handler(*args, **kwargs):
        body = cherrypy.request._json_inner_handler(*args, **kwargs)
        return json.dumps(body).encode('utf8') if cherrypy.response.headers['content-type'] == content_type else body

    cherrypy.lib.jsontools.json_out(content_type, handler=handler, **kwargs)


@cherrypy.tools.register('on_start_resource')
def allow(methods=None, paths=(), **kwargs):
    """Only allow specified methods."""
    handler = cherrypy.request.handler
    if paths and hasattr(handler, 'args'):
        with HTTPError(IndexError, http.client.NOT_FOUND):
            methods = paths[len(handler.args)]
    cherrypy.lib.cptools.allow(methods, **kwargs)


@cherrypy.tools.register('before_finalize')
def timer():
    """Return response time in headers."""
    response = cherrypy.serving.response
    response.headers['x-response-time'] = time.time() - response.time


@cherrypy.tools.register('on_start_resource')
def validate(etag=True, last_modified=False, max_age=None, expires=None):
    """Return and validate caching headers.

    :param etag: return weak entity tag header based on index version and validate if-match headers
    :param last_modified: return last-modified header based on index timestamp and validate if-modified headers
    :param max_age: return cache-control max-age and age headers based on last update timestamp
    :param expires: return expires header offset from last update timestamp
    """
    root = cherrypy.request.app.root
    headers = cherrypy.response.headers
    if etag:
        headers['etag'] = root.etag
        cherrypy.lib.cptools.validate_etags()
    if last_modified:
        headers['last-modified'] = cherrypy.lib.httputil.HTTPDate(root.searcher.timestamp)
        cherrypy.lib.cptools.validate_since()
    if max_age is not None:
        headers['age'] = int(time.time() - root.updated)
        headers['cache-control'] = 'max-age={}'.format(max_age)
    if expires is not None:
        headers['expires'] = cherrypy.lib.httputil.HTTPDate(expires + root.updated)


def multi(value):
    return value and value.split(',')


class parse:
    """Parameter parsing."""

    @staticmethod
    def q(searcher, q, **options):
        options = {key.partition('.')[-1]: options[key] for key in options if key.startswith('q.')}
        field = options.pop('field', [])
        fields = [field] if isinstance(field, str) else field
        fields = [name.partition('^')[::2] for name in fields]
        if any(boost for name, boost in fields):
            field = {name: float(boost or 1.0) for name, boost in fields}
        elif isinstance(field, str):
            ((field, boost),) = fields
        else:
            field = [name for name, boost in fields] or ''
        if 'type' in options:
            with HTTPError(AttributeError):
                return getattr(engine.Query, options['type'])(field, q)
        for key in set(options) - {'op', 'version'}:
            with HTTPError(ValueError):
                options[key] = json.loads(options[key])
        if q is not None:
            with HTTPError(lucene.JavaError):
                return searcher.parse(q, field=field, **options)

    @staticmethod
    def fields(searcher, fields=None, **options):
        if fields is not None:
            fields = dict.fromkeys(fields)
        multi = set(options.get('fields.multi', ()))
        docvalues = dict(parse.docvalues(searcher, field) for field in options.get('fields.docvalues', ()))
        return fields, multi, docvalues

    @staticmethod
    def docvalues(searcher, field):
        name, type = field.split(':') if ':' in field else (field, '')
        with HTTPError(AttributeError):
            return name, searcher.docvalues(name, getattr(__builtins__, type, None))


def json_error(version, **body):
    """Transform errors into json format."""
    tool = cherrypy.request.toolmaps['tools'].get('json_out', {})
    cherrypy.response.headers['content-type'] = tool.get('content_type', 'application/json')
    return json.dumps(body).encode('utf8')


def attach_thread(id=None):
    """Attach current cherrypy worker thread to lucene VM."""
    lucene.getVMEnv().attachCurrentThread()


class Autoreloader(cherrypy.process.plugins.Autoreloader):
    """Autoreload monitor compatible with lucene VM."""

    def run(self):
        attach_thread()
        cherrypy.process.plugins.Autoreloader.run(self)


class AttachedMonitor(cherrypy.process.plugins.Monitor):
    """Periodically run a callback function in an attached thread."""

    def __init__(self, bus, callback, frequency=cherrypy.process.plugins.Monitor.frequency):
        def run():
            attach_thread()
            callback()

        cherrypy.process.plugins.Monitor.__init__(self, bus, run, frequency)

    def subscribe(self):
        cherrypy.process.plugins.Monitor.subscribe(self)
        if cherrypy.engine.state == cherrypy.engine.states.STARTED:
            self.start()

    def unsubscribe(self):
        cherrypy.process.plugins.Monitor.unsubscribe(self)
        self.thread.cancel()


class WebSearcher(object):
    """Dispatch root with a delegated Searcher.

    :param urls: ordered hosts to synchronize with
    """

    _cp_config = {
        'tools.gzip.on': True,
        'tools.gzip.mime_types': ['text/html', 'text/plain', 'application/json'],
        'tools.accept.on': True,
        'tools.accept.media': 'application/json',
        'tools.json_in.on': True,
        'tools.json_out.on': True,
        'tools.allow.on': True,
        'tools.timer.on': True,
        'tools.validate.on': True,
        'error_page.default': json_error,
    }

    def __init__(self, *directories, **kwargs):
        self.urls = collections.deque(kwargs.pop('urls', ()))
        if self.urls:
            engine.IndexWriter(*directories).close()
        self.searcher = engine.MultiSearcher(directories, **kwargs) if len(directories) > 1 else engine.IndexSearcher(*directories, **kwargs)
        self.updated = time.time()
        self.query_map = {}

    @classmethod
    def new(cls, *args, **kwargs):
        """Return new uninitialized root which can be mounted on dispatch tree before VM initialization."""
        self = object.__new__(cls)
        self.args, self.kwargs = args, kwargs
        return self

    def close(self):
        self.searcher.close()

    @property
    def etag(self):
        return 'W/"{}"'.format(self.searcher.version)

    def sync(self, url):
        """Sync with remote index."""
        directory = self.searcher.path
        resource = clients.Resource(url, headers={'if-none-match': self.etag})
        response = resource.client.put('update/snapshot')
        if response.status_code in (http.client.PRECONDITION_FAILED, http.client.METHOD_NOT_ALLOWED):
            return []
        response.raise_for_status()
        names = sorted(set(response.json()).difference(os.listdir(directory)))
        resource /= response.headers['location']
        try:
            for name in names:
                with open(os.path.join(directory, name), 'wb') as file:
                    resource.download(file, name)
        finally:
            resource.delete()
        return names

    @cherrypy.expose
    @cherrypy.tools.json_in(process_body=dict)
    @cherrypy.tools.allow(methods=['GET', 'POST'])
    def index(self, url=''):
        """Return index information and synchronize with remote index.

        **GET, POST** /[index]
            Return a mapping of the directory to the document count.
            Add new segments from remote host.

            {"url": *string*}

            :return: {*string*: *int*,... }
        """
        if cherrypy.request.method == 'POST':
            self.sync(url)
            cherrypy.response.status = int(http.client.ACCEPTED)
        if isinstance(self.searcher, engine.MultiSearcher):
            return {reader.directory().toString(): reader.numDocs() for reader in self.searcher.indexReaders}
        return {self.searcher.directory.toString(): len(self.searcher)}

    @cherrypy.expose
    @cherrypy.tools.json_in(process_body=dict)
    @cherrypy.tools.allow(methods=['POST'])
    def update(self, **caches):
        """Refresh index version.

        **POST** /update
            Reopen searcher, optionally reloading caches, and return document count.

            {"spellcheckers": true,... }

            .. versionchanged:: 1.2 request body is an object instead of an array

            :return: *int*
        """
        names = ()
        while self.urls:
            url = self.urls[0]
            try:
                names = self.sync(url)
                break
            except IOError:
                with contextlib.suppress(ValueError):
                    self.urls.remove(url)
        self.searcher = self.searcher.reopen(**caches)
        self.updated = time.time()
        if names:
            engine.IndexWriter(self.searcher.directory).close()
        if not self.urls and hasattr(self, 'fields'):
            other = WebIndexer(self.searcher.directory, analyzer=self.searcher.analyzer)
            other.indexer.shared, other.indexer.fields = self.searcher.shared, self.fields
            (app,) = (app for app in cherrypy.tree.apps.values() if app.root is self)
            mount(other, app=app, autoupdate=getattr(self, 'autoupdate', 0))
        return len(self.searcher)

    @cherrypy.expose
    @cherrypy.tools.params()
    def docs(self, name=None, value='', **options):
        """Return ids or documents.

        **GET** /docs
            Return array of doc ids.

            :return: [*int*,... ]

        **GET** /docs/[*int*\|\ *chars*/*chars*]?
            Return document mapping from id or unique name and value.

            &fields=\ *chars*,... &fields.multi=\ *chars*,... &fields.docvalues=\ *chars*\ [:*chars*],...
                optionally select stored, multi-valued, and docvalues

            &fields.vector=\ *chars*,... &fields.vector.counts=\ *chars*,...
                optionally select term vectors with term counts

            :return: {*string*: null|\ *string*\|\ *number*\|\ *array*\|\ *object*,... }
        """
        searcher = self.searcher
        if not name:
            return list(searcher)
        with HTTPError(ValueError, http.client.NOT_FOUND):
            (id,) = searcher.docs(name, value) if value else [int(name)]
        fields, multi, docvalues = parse.fields(searcher, **options)
        with HTTPError(lucene.JavaError, http.client.NOT_FOUND):
            doc = searcher[id] if fields is None else searcher.get(id, *itertools.chain(fields, multi))
        result = doc.dict(*multi, **(fields or {}))
        with HTTPError(TypeError):
            result.update((name, docvalues[name][id]) for name in docvalues)
        result.update((field, list(searcher.termvector(id, field))) for field in options.get('fields.vector', ()))
        result.update((field, dict(searcher.termvector(id, field, counts=True))) for field in options.get('fields.vector.counts', ()))
        return result

    docs.__annotations__.update(dict.fromkeys(['fields', 'fields.multi', 'fields.docvalues', 'fields.vector', 'fields.vector.counts'], multi))

    @cherrypy.expose
    @cherrypy.tools.params()
    def search(
        self,
        q=None,
        count: int = None,
        start: int = 0,
        fields: multi = None,
        sort: multi = None,
        facets: multi = '',
        group='',
        hl: multi = '',
        mlt: int = None,
        timeout: float = None,
        **options
    ):
        """Run query and return documents.

        .. versionchanged:: 2.3 maxscore option and result removed
        .. versionchanged:: 2.4 __keys__ renamed to __sortkeys__

        **GET** /search?
            Return array of document objects and total doc count.

            &q=\ *chars*\ &q.type=[term|prefix|wildcard]&q.spellcheck=true&q.\ *chars*\ =...,
                query, optional type to skip parsing, spellcheck, and parser settings: q.field, q.op,...

            &count=\ *int*\ &start=0&count.min=1000
                maximum number of docs to return and offset to start at

            &fields=\ *chars*,... &fields.multi=\ *chars*,... &fields.docvalues=\ *chars*\ [:*chars*],...
                only include selected stored fields; multi-valued fields returned in an array; docvalues fields

            &sort=\ [-]\ *chars*\ [:*chars*],... &sort.scores
                | field name, optional type, minus sign indicates descending
                | optionally score docs

            &facets=\ *chars*,... &facets.count=\ *int*\&facets.min=0
                | include facet counts for given field names
                | optional maximum number of most populated facet values per field, and minimum count to return

            &group=\ *chars*\ [:*chars*]&group.count=1
                | group documents by field value with optional type, up to given maximum count

            .. versionchanged:: 1.6 grouping searches use count and start options

            &hl=\ *chars*,... &hl.count=1
                | stored fields to return highlighted
                | optional maximum fragment count

            &mlt=\ *int*\ &mlt.fields=\ *chars*,... &mlt.\ *chars*\ =...,
                | doc index (or id without a query) to find MoreLikeThis
                | optional document fields to match
                | optional MoreLikeThis settings: mlt.minTermFreq, mlt.minDocFreq,...

            &timeout=\ *number*
                timeout search after elapsed number of seconds

            :return:
                | {
                | "query": *string*\|null,
                | "count": *int*\|null,
                | "docs": [{"__id__": *int*, "__score__": *number*, "__sortkeys__": *array*,
                    "__highlights__": {*string*: *array*,... }, *string*: *value*,... },... ],
                | "facets": {*string*: {*string*: *int*,... },... },
                | "groups": [{"count": *int*, "value": *value*, "docs": [*object*,... ]},... ]
                | }
        """
        searcher = self.searcher
        if sort is not None:
            sort = (re.match('(-?)(\w+):?(\w*)', field).groups() for field in sort)
            with HTTPError(AttributeError):
                sort = [searcher.sortfield(name, getattr(__builtins__, type, None), (reverse == '-')) for reverse, name, type in sort]
        q = parse.q(searcher, q, **options)
        if mlt is not None:
            if q is not None:
                (mlt,) = searcher.search(q, count=mlt + 1, sort=sort)[mlt:].ids
            mltfields = options.pop('mlt.fields', ())
            with HTTPError(ValueError):
                attrs = {key.partition('.')[-1]: json.loads(options[key]) for key in options if key.startswith('mlt.')}
            q = searcher.morelikethis(mlt, *mltfields, analyzer=searcher.analyzer, **attrs)
        if count is not None:
            count += start
        if count == 0:
            start = count = 1
        scores = 'sort.scores' in options
        gcount = options.get('group.count', 1)
        mincount = options.get('count.min', 1000)
        if ':' in group:
            hits = searcher.search(q, sort=sort, timeout=timeout, scores=scores, mincount=mincount)
            name, docvalues = parse.docvalues(searcher, group)
            with HTTPError(TypeError):
                groups = hits.groupby(docvalues.select(hits.ids).__getitem__, count=count, docs=gcount)
            groups.groupdocs = groups.groupdocs[start:]
        elif group:
            groups = searcher.groupby(group, q, count, start=start, sort=sort, groupDocsLimit=gcount)
        else:
            hits = searcher.search(q, sort=sort, count=count, timeout=timeout, scores=scores, mincount=mincount)
            groups = engine.documents.Groups(searcher, [hits[start:]], hits.count)
        result = {'query': q and str(q), 'count': groups.count}
        fields, multi, docvalues = parse.fields(searcher, fields, **options)
        if fields is None:
            fields = {}
        else:
            groups.select(*itertools.chain(fields, multi))
        hl = dict.fromkeys(hl, options.get('hl.count', 1))
        result['groups'] = []
        for hits in groups:
            docs = []
            highlights = hits.highlights(q, **hl) if hl else ([{}] * len(hits))
            for hit, highlight in zip(hits, highlights):
                doc = hit.dict(*multi, **fields)
                with HTTPError(TypeError):
                    doc.update((name, docvalues[name][hit.id]) for name in docvalues)
                if highlight:
                    doc['__highlights__'] = highlight
                docs.append(doc)
            result['groups'].append({'docs': docs, 'count': hits.count, 'value': getattr(hits, 'value', None)})
        if not group:
            result['docs'] = result.pop('groups')[0]['docs']
        q = q or engine.Query.alldocs()
        if facets:
            query_map = {facet: self.query_map[facet] for facet in set(facets).intersection(self.query_map)}
            facets = result['facets'] = searcher.facets(q, *set(facets).difference(query_map), **query_map)
            for counts in facets.values():
                counts.pop(None, None)
            if 'facets.min' in options:
                for name, counts in facets.items():
                    facets[name] = {term: count for term, count in counts.items() if count >= options['facets.min']}
            if 'facets.count' in options:
                for name, counts in facets.items():
                    facets[name] = {term: counts[term] for term in heapq.nlargest(options['facets.count'], counts, key=counts.__getitem__)}
        return result

    search.__annotations__.update(
        {
            'fields.multi': multi,
            'fields.docvalues': multi,
            'facets.count': int,
            'facets.min': int,
            'group.count': int,
            'hl.count': int,
            'mlt.fields': multi,
            'count.min': int,
        }
    )

    @cherrypy.expose
    @cherrypy.tools.params()
    def terms(self, name='', value='*', *path, count: int = 0):
        """Return data about indexed terms.

        **GET** /terms?
            Return field names.

            :return: [*string*,... ]

        **GET** /terms/*chars*
            Return term values for given field name.

            :return: [*string*,... ]

        **GET** /terms/*chars*/*chars*\[\*\|:*chars*\|~[\ *int*\]]
            Return term values (prefix, slices, or fuzzy terms) for given field name.

            :return: [*string*,... ]

        **GET** /terms/*chars*/*chars*\[\*\|~[\ *int*\]\]?count=\ *int*
            Return spellchecked term values ordered by decreasing document frequency.
            Prefixes (*) are optimized to be suitable for real-time query suggestions; all terms are cached.

            :return: [*string*,... ]

        **GET** /terms/*chars*/*chars*
            Return document count for given term.

            :return: *int*

        **GET** /terms/*chars*/*chars*/docs
            Return document ids for given term.

            :return: [*int*,... ]

        **GET** /terms/*chars*/*chars*/docs/counts
            Return document ids and frequency counts for given term.

            :return: [[*int*, *int*],... ]

        **GET** /terms/*chars*/*chars*/docs/positions
            Return document ids and positions for given term.

            :return: [[*int*, [*int*,... ]],... ]
        """
        searcher = self.searcher
        if not name:
            return sorted(fieldinfo.name for fieldinfo in searcher.fieldinfos.values())
        if ':' in value:
            return list(searcher.terms(name, *value.split(':')))
        if value.endswith('*'):
            value = value.rstrip('*')
            if count:
                return searcher.complete(name, value, count)
            return list(searcher.terms(name, value))
        if '~' in value:
            with HTTPError(ValueError):
                value, distance = value.split('~')
                distance = int(distance or 2)
            if count:
                return searcher.suggest(name, value, count, maxEdits=distance)
            return list(searcher.terms(name, value, distance=distance))
        if not path:
            return searcher.count(name, value)
        if path[0] == 'docs':
            if path[1:] == ():
                return list(searcher.docs(name, value))
            if path[1:] == ('counts',):
                return list(searcher.docs(name, value, counts=True))
            if path[1:] == ('positions',):
                return list(searcher.positions(name, value))
        raise cherrypy.NotFound()

    @cherrypy.expose
    @cherrypy.tools.allow(paths=[('GET',), ('GET', 'POST'), ('GET', 'PUT', 'DELETE')])
    def queries(self, name='', value=''):
        """Match a document against registered queries.

        Queries are cached by a unique name and value, suitable for document indexing.

        .. versionadded:: 1.4

        **GET** /queries
            Return query set names.

            :return: [*string*,... ]

        **GET, POST** /queries/*chars*
            Return query values and scores which match given document.

            {*string*: *string*,... }

            :return: {*string*: *number*,... }

        **GET, PUT, DELETE** /queries/*chars*/*chars*
            Return, create, or delete a registered query.

            *string*

            :return: *string*
        """
        request = cherrypy.serving.request
        if not name:
            return sorted(self.query_map)
        if not value:
            if request.method == 'GET':
                request.body.process()
            with HTTPError(KeyError, http.client.NOT_FOUND):
                queries = self.query_map[name]
            scores = self.searcher.match(getattr(request, 'json', {}), *queries.values())
            return dict(zip(queries, scores))
        if request.method == 'DELETE':
            return str(self.query_map.get(name, {}).pop(value, '')) or None
        if request.method == 'PUT':
            queries = self.query_map.setdefault(name, {})
            if value not in queries:
                cherrypy.response.status = int(http.client.CREATED)
            queries[value] = self.searcher.parse(request.json)
        with HTTPError(KeyError, http.client.NOT_FOUND):
            return str(self.query_map[name][value])


class WebIndexer(WebSearcher):
    """Dispatch root with a delegated Indexer, exposing write methods."""

    def __init__(self, *args, **kwargs):
        self.indexer = engine.Indexer(*args, **kwargs)
        self.updated = time.time()
        self.query_map = {}

    @property
    def searcher(self):
        return self.indexer.indexSearcher

    def close(self):
        self.indexer.close()
        super().close()

    def refresh(self):
        if self.indexer.nrt:
            self.indexer.refresh()
            self.updated = time.time()
        else:
            cherrypy.response.status = int(http.client.ACCEPTED)

    @cherrypy.expose
    @cherrypy.tools.allow(methods=['GET', 'POST'])
    def index(self):
        """Add indexes.  See :meth:`WebSearcher.index` for GET method.

        **POST** /[index]
            Add indexes without optimization.

            [*string*,... ]
        """
        request = cherrypy.serving.request
        if request.method == 'POST':
            for directory in getattr(request, 'json', ()):
                self.indexer += directory
            self.refresh()
        return {str(self.indexer.directory): len(self.indexer)}

    @cherrypy.expose
    @cherrypy.tools.json_in(process_body=dict)
    @cherrypy.tools.allow(paths=[('POST',), ('GET', 'PUT', 'DELETE'), ('GET',)])
    def update(self, id='', name='', **options):
        """Commit index changes and refresh index version.

        **POST** /update
            Commit write operations and return document count.  See :meth:`WebSearcher.update` for caching options.

            {"merge": true|\ *int*,... }

            .. versionchanged:: 1.2 request body is an object instead of an array

            :return: *int*

        **GET, PUT, DELETE** /update/[snapshot|\ *int*]
            Verify, create, or release unique snapshot of current index commit and return array of referenced filenames.

            .. versionchanged:: 1.4 lucene identifies snapshots by commit generation;  use location header

            :return: [*string*,... ]

        **GET** /update/*int*/*chars*
            Download index file corresponding to snapshot id and filename.
        """
        if not id:
            self.indexer.commit(**options)
            self.updated = time.time()
            return len(self.indexer)
        method = cherrypy.request.method
        response = cherrypy.serving.response
        if method == 'PUT':
            if id != 'snapshot':
                raise cherrypy.NotFound()
            commit = self.indexer.policy.snapshot()
            response.status = int(http.client.CREATED)
            response.headers['location'] = cherrypy.url('/update/{0:d}'.format(commit.generation), relative='server')
        else:
            with HTTPError((ValueError, AssertionError), http.client.NOT_FOUND):
                commit = self.indexer.policy.getIndexCommit(int(id))
                assert commit is not None, 'commit not snapshotted'
            if method == 'DELETE':
                self.indexer.policy.release(commit)
        if not name:
            return list(commit.fileNames)
        with HTTPError((TypeError, AssertionError), http.client.NOT_FOUND):
            directory = self.searcher.path
            assert name in commit.fileNames, 'file not referenced in commit'
        return cherrypy.lib.static.serve_download(os.path.join(directory, name))

    @cherrypy.expose
    @cherrypy.tools.allow(paths=[('GET', 'POST'), ('GET',), ('GET', 'PUT', 'DELETE', 'PATCH')])
    def docs(self, name=None, value='', **options):
        """Add or return documents.  See :meth:`WebSearcher.docs` for GET method.

        **POST** /docs
            Add documents to index.

            [{*string*: *string*\|\ *number*\|\ *array*,... },... ]

        **PUT, DELETE** /docs/*chars*/*chars*
            Set or delete document.  Unique term should be indexed and is added to the new document.

            {*string*: *string*\|\ *number*\|\ *array*,... }
        """
        request = cherrypy.serving.request
        if request.method in ('GET', 'HEAD'):
            return super().docs(name, value, **options)
        if request.method == 'DELETE':
            self.indexer.delete(name, value)
        elif request.method == 'POST':
            for doc in getattr(request, 'json', ()):
                self.indexer.add(doc)
        else:
            doc = getattr(request, 'json', {})
            with HTTPError((KeyError, AssertionError), http.client.CONFLICT):
                assert self.indexer.fields[name].indexed, 'unique field must be indexed'
            if request.method == 'PUT':
                with HTTPError(AssertionError):
                    assert doc.setdefault(name, value) == value, 'multiple values for unique field'
            else:
                with HTTPError((KeyError, AssertionError), http.client.CONFLICT):
                    assert all(self.indexer.fields[name].docvalues for name in doc)
            self.indexer.update(name, value, doc)
        self.refresh()

    docs._cp_config.update(WebSearcher.docs._cp_config)
    docs._cp_config['request.methods_with_bodies'] = ('POST', 'PUT', 'PATCH')
    docs.__annotations__.update(WebSearcher.docs.__annotations__)

    @cherrypy.expose
    @cherrypy.tools.allow(methods=['GET', 'DELETE'])
    def search(self, q=None, **options):
        """Run or delete a query.  See :meth:`WebSearcher.search` for GET method.

        **DELETE** /search?q=\ *chars*
            Delete documents which match query.
        """
        if cherrypy.request.method != 'DELETE':
            return super().search(q, **options)
        if q is None:
            self.indexer.deleteAll()
        else:
            self.indexer.delete(parse.q(self.searcher, q, **options))
        self.refresh()

    search._cp_config.update(WebSearcher.search._cp_config)
    search.__annotations__.update(WebSearcher.search.__annotations__)

    @cherrypy.expose
    @cherrypy.tools.json_in(process_body=dict)
    @cherrypy.tools.allow(paths=[('GET',), ('GET', 'PUT')])
    @cherrypy.tools.validate(on=False)
    def fields(self, name='', **settings):
        """Return or store a field's settings.

        **GET** /fields
            Return known field names.

            :return: [*string*,... ]

        **GET, PUT** /fields/*chars*
            Set and return settings for given field name.

            {"stored"|"indexOptions"\|...: *string*\|true|false,... }

            .. versionchanged:: 1.6 lucene FieldType attributes used as settings

            :return: {"stored"|"indexOptions"\|...: *string*\|true|false,... }
        """
        if not name:
            return sorted(self.indexer.fields)
        if cherrypy.request.method == 'PUT':
            if name not in self.indexer.fields:
                cherrypy.response.status = int(http.client.CREATED)
            with HTTPError(AttributeError):
                self.indexer.set(name, **settings)
        with HTTPError(KeyError, http.client.NOT_FOUND):
            return self.indexer.fields[name].settings


def init(vmargs='-Xrs,-Djava.awt.headless=true', **kwargs):
    """Callback to initialize VM and app roots after daemonizing."""
    assert lucene.getVMEnv() or lucene.initVM(vmargs=vmargs, **kwargs)
    for app in cherrypy.tree.apps.values():
        if isinstance(app.root, WebSearcher):
            app.root.__init__(*app.root.__dict__.pop('args'), **app.root.__dict__.pop('kwargs'))


def mount(root, path='', config=None, autoupdate=0, app=None):
    """Attach root and subscribe to plugins.

    :param root,path,config: see cherrypy.tree.mount
    :param autoupdate: see command-line options
    :param app: optionally replace root on existing app
    """
    if app is None:
        app = cherrypy.tree.mount(root, path, config)
    else:
        cherrypy.engine.unsubscribe('stop', app.root.close)
        if hasattr(app.root, 'monitor'):
            app.root.monitor.unsubscribe()
        app.root = root
    cherrypy.engine.subscribe('stop', root.close)
    if autoupdate:
        root.monitor = AttachedMonitor(cherrypy.engine, root.update, autoupdate)
        root.monitor.subscribe()
    return app


def start(root=None, path='', config=None, pidfile='', daemonize=False, autoreload=0, autoupdate=0, callback=None):
    """Attach root, subscribe to plugins, and start server.

    :param root,path,config: see cherrypy.quickstart
    :param pidfile,daemonize,autoreload,autoupdate: see command-line options
    :param callback: optional callback function scheduled after daemonizing
    """
    cherrypy.engine.subscribe('start_thread', attach_thread)
    cherrypy.config['engine.autoreload.on'] = False
    if pidfile:
        cherrypy.process.plugins.PIDFile(cherrypy.engine, os.path.abspath(pidfile)).subscribe()
    if daemonize:
        cherrypy.config['log.screen'] = False
        cherrypy.process.plugins.Daemonizer(cherrypy.engine).subscribe()
    if autoreload:
        reloader = Autoreloader(cherrypy.engine, autoreload, match='lupyne.*')
        reloader.files.add(__file__)
        reloader.subscribe()
    if callback:
        priority = (cherrypy.process.plugins.Daemonizer.start.priority + cherrypy.process.plugins.Monitor.start.priority) // 2
        cherrypy.engine.subscribe('start', callback, priority)
    if root is not None:
        mount(root, path, config, autoupdate)
    cherrypy.quickstart(cherrypy.tree.apps.get(path), path, config)


parser = argparse.ArgumentParser(description='Restful json cherrypy server.', prog='lupyne.server')
parser.add_argument('directories', nargs='*', metavar='directory', help='index directories')
parser.add_argument('-r', '--read-only', action='store_true', help='expose only read methods; no write lock')
parser.add_argument('-c', '--config', help='optional configuration file or json object of global params')
parser.add_argument('-p', '--pidfile', metavar='FILE', help='store the process id in the given file')
parser.add_argument('-d', '--daemonize', action='store_true', help='run the server as a daemon')
parser.add_argument('--autoreload', type=float, metavar='SECONDS', help='automatically reload modules; replacement for engine.autoreload')
parser.add_argument('--autoupdate', type=float, metavar='SECONDS', help='automatically update index version and commit any changes')
parser.add_argument('--autosync', metavar='URL,...', help='automatically synchronize searcher with remote hosts and update')
parser.add_argument('--real-time', action='store_true', help='search in real-time without committing')

if __name__ == '__main__':
    args = parser.parse_args()
    read_only = args.read_only or args.autosync or len(args.directories) > 1
    kwargs = {'nrt': True} if args.real_time else {}
    if read_only and (args.real_time or not args.directories):
        parser.error('incompatible read/write options')
    if args.autosync:
        kwargs['urls'] = args.autosync.split(',')
        if not (args.autoupdate and len(args.directories) == 1):
            parser.error('autosync requires autoupdate and a single directory')
        warnings.warn('autosync is not recommended for production usage')
    if args.config and not os.path.exists(args.config):
        args.config = {'global': json.loads(args.config)}
    cls = WebSearcher if read_only else WebIndexer
    root = cls.new(*map(os.path.abspath, args.directories), **kwargs)
    del args.directories, args.read_only, args.autosync, args.real_time
    start(root, callback=init, **args.__dict__)



import abc
import bisect
import contextlib
import heapq
import itertools
import types
import lucene
import six
from java.lang import Double, Float, Number, Object
from org.apache.lucene import analysis, util

try:
    from typing import Iterable
except ImportError:  # pragma: no cover
    from collections import Iterable
long = int if six.PY3 else long  # noqa


class Atomic(six.with_metaclass(abc.ABCMeta)):
    """Abstract base class to distinguish singleton values from other iterables."""

    @classmethod
    def __subclasshook__(cls, other):
        return not issubclass(other, Iterable) or NotImplemented


for cls in six.string_types + (analysis.TokenStream, lucene.JArray_byte):
    Atomic.register(cls)


class method(staticmethod):
    """Backport of Python 3's unbound methods."""

    def __get__(self, instance, owner):
        return self.__func__ if instance is None else types.MethodType(self.__func__, instance)


class SpellChecker(dict):
    """Correct spellings and suggest words for queries.

    Supply a vocabulary mapping words to (reverse) sort keys, such as document frequencies.
    """

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.words = sorted(self)
        alphabet = ''.join(set(itertools.chain.from_iterable(self)))
        self.suffix = alphabet and max(alphabet) * max(map(len, self))

    def complete(self, prefix, count=None):
        """Return ordered suggested words for prefix."""
        start = bisect.bisect_left(self.words, prefix)
        stop = bisect.bisect_right(self.words, prefix + self.suffix, start)
        words = self.words[start:stop]
        if count is not None and count < len(words):
            return heapq.nlargest(count, words, key=self.__getitem__)
        words.sort(key=self.__getitem__, reverse=True)
        return words


@contextlib.contextmanager
def suppress(exception):
    """Suppress specific lucene exception."""
    try:
        yield
    except lucene.JavaError as exc:
        if not exception.instance_(exc.getJavaException()):
            raise


def convert(value):
    """Return python object from java Object."""
    if util.BytesRef.instance_(value):
        return util.BytesRef.cast_(value).utf8ToString()
    if not Number.instance_(value):
        return value.toString() if Object.instance_(value) else value
    value = Number.cast_(value)
    return value.doubleValue() if Float.instance_(value) or Double.instance_(value) else int(value.longValue())



import contextlib
import itertools
import operator
import os
from functools import partial
import lucene
from java.io import File, IOException, StringReader
from java.util import Arrays, HashSet
from org.apache.lucene import analysis, document, index, queries, search, store, util
from org.apache.lucene.search import spell, uhighlight
from six import string_types
from six.moves import filter, map, range, zip
from .analyzers import Analyzer
from .queries import Query, DocValues, SpellParser
from .documents import Field, Document, Hits, GroupingSearch
from .utils import long, suppress, Atomic, SpellChecker


class closing(set):
    """Manage lifespan of registered objects, similar to contextlib.closing."""

    def __del__(self):
        for obj in self:
            obj.close()

    def analyzer(self, analyzer):
        if analyzer is None:
            analyzer = analysis.standard.StandardAnalyzer()
            self.add(analyzer)
        return analyzer

    def directory(self, directory):
        if directory is None:
            directory = store.RAMDirectory()
            self.add(directory)
        elif isinstance(directory, string_types):
            directory = store.FSDirectory.open(File(directory).toPath())
            self.add(directory)
        return directory

    def reader(self, reader):
        if isinstance(reader, index.IndexReader):
            reader.incRef()
        elif isinstance(reader, index.IndexWriter):
            reader = index.DirectoryReader.open(reader)
        elif isinstance(reader, Atomic):
            reader = index.DirectoryReader.open(self.directory(reader))
        else:
            reader = index.MultiReader(list(map(self.reader, reader)))
        return reader

    @classmethod
    @contextlib.contextmanager
    def store(cls, directory):
        self = cls()
        yield self.directory(directory)


def copy(commit, dest):
    """Copy the index commit to the destination directory.

    Optimized to use hard links if the destination is a file system path.
    """
    if isinstance(dest, store.Directory):
        for filename in commit.fileNames:
            dest.copyFrom(commit.directory, filename, filename, store.IOContext.DEFAULT)
    else:
        src = IndexSearcher.path.fget(commit)
        os.path.isdir(dest) or os.makedirs(dest)
        for filename in commit.fileNames:
            paths = os.path.join(src, filename), os.path.join(dest, filename)
            try:
                os.link(*paths)
            except OSError:
                if not os.path.samefile(*paths):
                    raise


class IndexReader(object):
    """Delegated lucene IndexReader, with a mapping interface of ids to document objects.

    :param reader: lucene IndexReader
    """

    def __init__(self, reader):
        self.indexReader = reader

    def __getattr__(self, name):
        if name == 'indexReader':
            raise AttributeError(name)
        return getattr(index.DirectoryReader.cast_(self.indexReader), name)

    def __len__(self):
        return self.numDocs()

    def __contains__(self, id):
        bits = self.bits
        return (0 <= id < self.maxDoc()) and (not bits or bits.get(id))

    def __iter__(self):
        ids = range(self.maxDoc())
        bits = self.bits
        return filter(bits.get, ids) if bits else iter(ids)

    @property
    def bits(self):
        return index.MultiBits.getLiveDocs(self.indexReader)

    @property
    def directory(self):
        """reader's lucene Directory"""
        return self.__getattr__('directory')()

    @property
    def path(self):
        """FSDirectory path"""
        return str(store.FSDirectory.cast_(self.directory).directory)

    @property
    def timestamp(self):
        """timestamp of reader's last commit"""
        return File(self.path, self.indexCommit.segmentsFileName).lastModified() * 0.001

    @property
    def readers(self):
        """segment readers"""
        return (index.SegmentReader.cast_(context.reader()) for context in self.leaves())

    @property
    def segments(self):
        """segment filenames with document counts"""
        return {reader.segmentName: reader.numDocs() for reader in self.readers}

    @property
    def fieldinfos(self):
        """mapping of field names to lucene FieldInfos"""
        fieldinfos = index.FieldInfos.getMergedFieldInfos(self.indexReader)
        return {fieldinfo.name: fieldinfo for fieldinfo in fieldinfos.iterator()}

    def suggest(self, name, value, count=1, **attrs):
        """Return spelling suggestions from DirectSpellChecker.

        :param name: field name
        :param value: term
        :param count: maximum number of suggestions
        :param attrs: DirectSpellChecker options
        """
        checker = spell.DirectSpellChecker()
        for attr in attrs:
            setattr(checker, attr, attrs[attr])
        return [word.string for word in checker.suggestSimilar(index.Term(name, value), count, self.indexReader)]

    def sortfield(self, name, type=None, reverse=False):
        """Return lucene SortField, deriving the the type from FieldInfos if necessary.

        :param name: field name
        :param type: int, float, or name compatible with SortField constants
        :param reverse: reverse flag used with sort
        """
        if type is None:
            type = self.fieldinfos[name].docValuesType.toString()
        type = Field.types.get(type, type).upper()
        return search.SortField(name, getattr(search.SortField.Type, type), reverse)

    def docvalues(self, name, type=None):
        """Return chained lucene DocValues, suitable for custom sorting or grouping.

        Note multi-valued DocValues aren't thread-safe and only supported ordered iteration.

        :param name: field name
        :param type: int or float for converting values
        """
        type = {int: int, float: util.NumericUtils.sortableLongToDouble}.get(type, util.BytesRef.utf8ToString)
        docValuesType = self.fieldinfos[name].docValuesType.toString().title().replace('_', '')
        method = getattr(index.MultiDocValues, 'get{}Values'.format(docValuesType))
        return getattr(DocValues, docValuesType)(method(self.indexReader, name), len(self), type)

    def copy(self, dest, query=None, exclude=None, merge=0):
        """Copy the index to the destination directory.

        Optimized to use hard links if the destination is a file system path.

        :param dest: destination directory path or lucene Directory
        :param query: optional lucene Query to select documents
        :param exclude: optional lucene Query to exclude documents
        :param merge: optionally merge into maximum number of segments
        """
        copy(self.indexCommit, dest)
        with IndexWriter(dest) as writer:
            if query:
                writer.delete(Query.alldocs() - query)
            if exclude:
                writer.delete(exclude)
            writer.commit()
            writer.forceMergeDeletes()
            if merge:
                writer.forceMerge(merge)
            return len(writer)

    def terms(self, name, value='', stop='', counts=False, distance=0, prefix=0):
        """Generate a slice of term values, optionally with frequency counts.

        :param name: field name
        :param value: term prefix, start value (given stop), or fuzzy value (given distance)
        :param stop: optional upper bound for range terms
        :param counts: include frequency counts
        :param distance: maximum edit distance for fuzzy terms
        :param prefix: prefix length for fuzzy terms
        """
        terms = index.MultiTerms.getTerms(self.indexReader, name)
        if not terms:
            return iter([])
        term, termsenum = index.Term(name, value), terms.iterator()
        if distance:
            terms = termsenum = search.FuzzyTermsEnum(terms, util.AttributeSource(), term, distance, prefix, False)
        else:
            termsenum.seekCeil(util.BytesRef(value))
            terms = itertools.chain([termsenum.term()], util.BytesRefIterator.cast_(termsenum))
        terms = map(operator.methodcaller('utf8ToString'), terms)
        predicate = partial(operator.gt, stop) if stop else operator.methodcaller('startswith', value)
        if not distance:
            terms = itertools.takewhile(predicate, terms)
        return ((term, termsenum.docFreq()) for term in terms) if counts else terms

    def docs(self, name, value, counts=False):
        """Generate doc ids which contain given term, optionally with frequency counts."""
        docsenum = index.MultiTerms.getTermPostingsEnum(self.indexReader, name, util.BytesRef(value))
        docs = iter(docsenum.nextDoc, index.PostingsEnum.NO_MORE_DOCS) if docsenum else ()
        return ((doc, docsenum.freq()) for doc in docs) if counts else iter(docs)

    def positions(self, name, value, payloads=False, offsets=False):
        """Generate doc ids and positions which contain given term, optionally with offsets, or only ones with payloads."""
        docsenum = index.MultiTerms.getTermPostingsEnum(self.indexReader, name, util.BytesRef(value))
        for doc in iter(docsenum.nextDoc, index.PostingsEnum.NO_MORE_DOCS) if docsenum else ():
            positions = (docsenum.nextPosition() for _ in range(docsenum.freq()))
            if payloads:
                positions = ((position, docsenum.payload.utf8ToString()) for position in positions if docsenum.payload)
            elif offsets:
                positions = ((docsenum.startOffset(), docsenum.endOffset()) for position in positions)
            yield doc, list(positions)

    def vector(self, id, field):
        terms = self.getTermVector(id, field)
        termsenum = terms.iterator() if terms else index.TermsEnum.EMPTY
        return termsenum, map(operator.methodcaller('utf8ToString'), util.BytesRefIterator.cast_(termsenum))

    def termvector(self, id, field, counts=False):
        """Generate terms for given doc id and field, optionally with frequency counts."""
        termsenum, terms = self.vector(id, field)
        return ((term, int(termsenum.totalTermFreq())) for term in terms) if counts else terms

    def positionvector(self, id, field, offsets=False):
        """Generate terms and positions for given doc id and field, optionally with character offsets."""
        termsenum, terms = self.vector(id, field)
        for term in terms:
            docsenum = termsenum.postings(None)
            assert 0 <= docsenum.nextDoc() < docsenum.NO_MORE_DOCS
            positions = (docsenum.nextPosition() for _ in range(docsenum.freq()))
            if offsets:
                positions = ((docsenum.startOffset(), docsenum.endOffset()) for _ in positions)
            yield term, list(positions)

    def morelikethis(self, doc, *fields, **attrs):
        """Return MoreLikeThis query for document.

        :param doc: document id or text
        :param fields: document fields to use, optional for termvectors
        :param attrs: additional attributes to set on the morelikethis object
        """
        mlt = queries.mlt.MoreLikeThis(self.indexReader)
        mlt.fieldNames = fields or None
        for name, value in attrs.items():
            setattr(mlt, name, value)
        return mlt.like(fields[0], StringReader(doc)) if isinstance(doc, string_types) else mlt.like(doc)


class IndexSearcher(search.IndexSearcher, IndexReader):
    """Inherited lucene IndexSearcher, with a mixed-in IndexReader.

    :param directory: directory path, lucene Directory, or lucene IndexReader
    :param analyzer: lucene Analyzer, default StandardAnalyzer
    """

    def __init__(self, directory, analyzer=None):
        self.shared = closing()
        search.IndexSearcher.__init__(self, self.shared.reader(directory))
        self.analyzer = self.shared.analyzer(analyzer)
        self.spellcheckers = {}

    @classmethod
    def load(cls, directory, analyzer=None):
        """Open `IndexSearcher`_ with a lucene RAMDirectory, loading index into memory."""
        with closing.store(directory) as directory:
            directory = store.RAMDirectory(directory, store.IOContext.DEFAULT)
        self = cls(directory, analyzer)
        self.shared.add(self.directory)
        return self

    def __del__(self):
        if hash(self):
            self.decRef()

    def openIfChanged(self):
        return index.DirectoryReader.openIfChanged(index.DirectoryReader.cast_(self.indexReader))

    def reopen(self, spellcheckers=False):
        """Return current `IndexSearcher`_, only creating a new one if necessary.

        :param spellcheckers: refresh cached :attr:`spellcheckers`
        """
        reader = self.openIfChanged()
        if reader is None:
            return self
        other = type(self)(reader, self.analyzer)
        other.decRef()
        other.shared = self.shared
        if spellcheckers:
            for field in self.spellcheckers:
                other.spellchecker(field)
        else:
            other.spellcheckers = dict(self.spellcheckers)
        return other

    def __getitem__(self, id):
        return Document(self.doc(id))

    def get(self, id, *fields):
        """Return `Document`_ with only selected fields loaded."""
        return Document(self.document(id, HashSet(Arrays.asList(fields))))

    def spans(self, query, positions=False):
        """Generate docs with occurrence counts for a span query.

        :param query: lucene SpanQuery
        :param positions: optionally include slice positions instead of counts
        """
        offset = 0
        weight = query.createWeight(self, search.ScoreMode.COMPLETE_NO_SCORES, 1.0)
        postings = search.spans.SpanWeight.Postings.POSITIONS
        for reader in self.readers:
            try:
                spans = weight.getSpans(reader.context, postings)
            except lucene.JavaError:  # EOF
                continue
            for doc in iter(spans.nextDoc, spans.NO_MORE_DOCS):
                starts = iter(spans.nextStartPosition, spans.NO_MORE_POSITIONS)
                if positions:
                    values = [(start, spans.endPosition()) for start in starts]
                else:
                    values = sum(1 for _ in starts)
                yield (doc + offset), values
            offset += reader.maxDoc()

    def parse(self, query, spellcheck=False, **kwargs):
        if isinstance(query, search.Query):
            return query
        if spellcheck:
            kwargs['parser'], kwargs['searcher'] = SpellParser, self
        return Analyzer.parse(self.analyzer, query, **kwargs)

    @property
    def highlighter(self):
        """lucene UnifiedHighlighter"""
        return uhighlight.UnifiedHighlighter(self, self.analyzer)

    def count(self, *query, **options):
        """Return number of hits for given query or term.

        :param query: :meth:`search` compatible query, or optimally a name and value
        :param options: additional :meth:`search` options
        """
        if len(query) > 1:
            return self.docFreq(index.Term(*query))
        query = self.parse(*query, **options) if query else Query.alldocs()
        return super(IndexSearcher, self).count(query)

    def collector(self, count=None, sort=None, reverse=False, scores=False, mincount=1000):
        if count is None:
            return search.CachingCollector.create(True, float('inf'))
        count = min(count, self.maxDoc() or 1)
        mincount = max(count, mincount)
        if sort is None:
            return search.TopScoreDocCollector.create(count, mincount)
        if isinstance(sort, string_types):
            sort = self.sortfield(sort, reverse=reverse)
        if not isinstance(sort, search.Sort):
            sort = search.Sort(sort)
        return search.TopFieldCollector.create(sort, count, mincount)

    def search(self, query=None, count=None, sort=None, reverse=False, scores=False, mincount=1000, timeout=None, **parser):
        """Run query and return `Hits`_.

        .. versionchanged:: 1.4 sort param for lucene only; use Hits.sorted with a callable
        .. versionchanged:: 2.3 maxscore option removed; use Hits.maxscore property

        :param query: query string or lucene Query
        :param count: maximum number of hits to retrieve
        :param sort: lucene Sort parameters
        :param reverse: reverse flag used with sort
        :param scores: compute scores for candidate results when sorting
        :param mincount: total hit count accuracy threshold
        :param timeout: stop search after elapsed number of seconds
        :param parser: :meth:`Analyzer.parse` options
        """
        query = Query.alldocs() if query is None else self.parse(query, **parser)
        cache = collector = self.collector(count, sort, reverse, scores, mincount)
        counter = search.TimeLimitingCollector.getGlobalCounter()
        results = collector if timeout is None else search.TimeLimitingCollector(collector, counter, long(timeout * 1000))
        with suppress(search.TimeLimitingCollector.TimeExceededException):
            search.IndexSearcher.search(self, query, results)
            timeout = None
        if isinstance(cache, search.CachingCollector):
            collector = search.TotalHitCountCollector()
            cache.replay(collector)
            count = collector.totalHits or 1
            collector = self.collector(count, sort, reverse, scores, count)
            cache.replay(collector)
        topdocs = collector.topDocs()
        if scores:
            search.TopFieldCollector.populateScores(topdocs.scoreDocs, self, query)
        return Hits(self, topdocs.scoreDocs, topdocs.totalHits)

    def facets(self, query, *fields, **query_map):
        """Return mapping of document counts for the intersection with each facet.

        .. versionchanged:: 1.6 filters are no longer implicitly cached, a `GroupingSearch`_ is used instead

        :param query: query string or lucene Query
        :param fields: field names for lucene GroupingSearch
        :param query_map: `{facet: {key: query, ...}, ...}` for intersected query counts
        """
        query = self.parse(query)
        counts = {field: self.groupby(field, query).facets for field in fields}
        for facet, queries in query_map.items():
            counts[facet] = {key: self.count(Query.all(query, queries[key])) for key in queries}
        return counts

    def groupby(self, field, query, count=None, start=0, **attrs):
        """Return `Hits`_ grouped by field using a `GroupingSearch`_."""
        return GroupingSearch(field, **attrs).search(self, self.parse(query), count, start)

    def spellchecker(self, field):
        """Return and cache spellchecker for given field."""
        try:
            return self.spellcheckers[field]
        except KeyError:
            return self.spellcheckers.setdefault(field, SpellChecker(self.terms(field, counts=True)))

    def complete(self, field, prefix, count=None):
        """Return ordered suggested words for prefix."""
        return self.spellchecker(field).complete(prefix, count)

    def match(self, document, *queries):
        """Generate scores for all queries against a given document mapping."""
        searcher = index.memory.MemoryIndex()
        for name, value in document.items():
            args = [self.analyzer] * isinstance(value, string_types)
            searcher.addField(name, value, *args)
        return (searcher.search(self.parse(query)) for query in queries)


class MultiSearcher(IndexSearcher):
    """IndexSearcher with underlying lucene MultiReader.

    :param reader: directory paths, Directories, IndexReaders, or a single MultiReader
    :param analyzer: lucene Analyzer, default StandardAnalyzer
    """

    def __init__(self, reader, analyzer=None):
        IndexSearcher.__init__(self, reader, analyzer)
        self.indexReaders = [index.DirectoryReader.cast_(context.reader()) for context in self.context.children()]
        self.version = sum(reader.version for reader in self.indexReaders)

    def __getattr__(self, name):
        return getattr(index.MultiReader.cast_(self.indexReader), name)

    def openIfChanged(self):
        readers = list(map(index.DirectoryReader.openIfChanged, self.indexReaders))
        if any(readers):
            return index.MultiReader([new or old.incRef() or old for new, old in zip(readers, self.indexReaders)])

    @property
    def timestamp(self):
        return max(IndexReader(reader).timestamp for reader in self.indexReaders)


class IndexWriter(index.IndexWriter):
    """Inherited lucene IndexWriter.

    Supports setting fields parameters explicitly, so documents can be represented as dictionaries.

    :param directory: directory path or lucene Directory, default RAMDirectory
    :param mode: file mode (rwa), except updating (+) is implied
    :param analyzer: lucene Analyzer, default StandardAnalyzer
    :param version: lucene Version argument passed to IndexWriterConfig, default is latest
    :param attrs: additional attributes to set on IndexWriterConfig
    """

    parse = IndexSearcher.__dict__['parse']

    def __init__(self, directory=None, mode='a', analyzer=None, version=None, **attrs):
        self.shared = closing()
        config = index.IndexWriterConfig() if analyzer is None else index.IndexWriterConfig(self.shared.analyzer(analyzer))
        config.openMode = index.IndexWriterConfig.OpenMode.values()['wra'.index(mode)]
        for name, value in attrs.items():
            setattr(config, name, value)
        self.policy = config.indexDeletionPolicy = index.SnapshotDeletionPolicy(config.indexDeletionPolicy)
        index.IndexWriter.__init__(self, self.shared.directory(directory), config)
        self.fields = {}

    def __del__(self):
        if hash(self):
            with suppress(IOException):
                self.close()

    def __len__(self):
        return self.docStats.numDocs

    @classmethod
    def check(cls, directory, fix=False):
        """Check and optionally fix unlocked index, returning lucene CheckIndex.Status."""
        with closing.store(directory) as directory:
            with contextlib.closing(index.CheckIndex(directory)) as checkindex:
                status = checkindex.checkIndex()
                if fix:
                    checkindex.exorciseIndex(status)
        return status

    def set(self, name, cls=Field, **settings):
        """Assign settings to field name and return the field.

        :param name: registered name of field
        :param cls: optional `Field`_ constructor
        :param settings: stored, indexed, etc. options compatible with `Field`_
        """
        field = self.fields[name] = cls(name, **settings)
        return field

    def document(self, items=(), **terms):
        """Return lucene Document from mapping of field names to one or multiple values."""
        doc = document.Document()
        for name, values in dict(items, **terms).items():
            if isinstance(values, Atomic):
                values = (values,)
            for field in self.fields[name].items(*values):
                doc.add(field)
        return doc

    def add(self, document=(), **terms):
        """Add :meth:`document` to index with optional boost."""
        self.addDocument(self.document(document, **terms))

    def update(self, name, value='', document=(), **terms):
        """Atomically delete documents which match given term and add the new :meth:`document`.

        .. versionchanged:: 1.7 update in-place if only DocValues are given
        """
        doc = self.document(document, **terms)
        term = index.Term(name, *[value] if value else doc.getValues(name))
        fields = list(doc.iterator())
        types = [Field.cast_(field.fieldType()) for field in fields]
        if any(ft.stored() or ft.indexOptions() != index.IndexOptions.NONE or Field.dimensions.fget(ft) for ft in types):
            self.updateDocument(term, doc)
        elif fields:
            self.updateDocValues(term, *fields)

    def delete(self, *query, **options):
        """Remove documents which match given query or term.

        :param query: :meth:`IndexSearcher.search` compatible query, or optimally a name and value
        :param options: additional :meth:`Analyzer.parse` options
        """
        parse = self.parse if len(query) == 1 else index.Term
        self.deleteDocuments(parse(*query, **options))

    def __iadd__(self, directory):
        """Add directory (or reader, searcher, writer) to index."""
        with closing.store(getattr(directory, 'directory', directory)) as directory:
            self.addIndexes([directory])
        return self

    @contextlib.contextmanager
    def snapshot(self):
        """Return context manager of an index commit snapshot.

        .. versionchanged:: 1.4 lucene identifies snapshots by commit generation
        """
        commit = self.policy.snapshot()
        try:
            yield commit
        finally:
            self.policy.release(commit)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if any(args):
            self.rollback()
        else:
            self.commit()
        self.close()


class Indexer(IndexWriter):
    """An all-purpose interface to an index.

    Creates an `IndexWriter`_ with a delegated `IndexSearcher`_.

    :param nrt: optionally use a near real-time searcher
    """

    def __init__(self, directory=None, mode='a', analyzer=None, version=None, nrt=False, **attrs):
        IndexWriter.__init__(self, directory, mode, analyzer, version, **attrs)
        IndexWriter.commit(self)
        self.nrt = nrt
        self.indexSearcher = IndexSearcher(self if nrt else self.directory, self.analyzer)

    def __getattr__(self, name):
        if name == 'indexSearcher':
            raise AttributeError(name)
        return getattr(self.indexSearcher, name)

    def __contains__(self, id):
        return id in self.indexSearcher

    def __iter__(self):
        return iter(self.indexSearcher)

    def __getitem__(self, id):
        return self.indexSearcher[id]

    def refresh(self, **caches):
        """Store refreshed searcher with :meth:`IndexSearcher.reopen` caches."""
        self.indexSearcher = self.indexSearcher.reopen(**caches)

    def commit(self, merge=False, **caches):
        """Commit writes and :meth:`refresh` searcher.

        :param merge: merge segments with deletes, or optionally specify maximum number of segments
        """
        IndexWriter.commit(self)
        if merge:
            if isinstance(merge, bool):
                self.forceMergeDeletes()
            else:
                self.forceMerge(merge)
            IndexWriter.commit(self)
        self.refresh(**caches)



from setuptools import setup
import lupyne

setup(
    name='lupyne',
    version=lupyne.__version__,
    description='Pythonic search engine based on PyLucene, including a standalone server based on CherryPy.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Aric Coady',
    author_email='aric.coady@gmail.com',
    url='https://github.com/coady/lupyne',
    project_urls={'Documentation': 'https://lupyne.surge.sh'},
    license='Apache Software License',
    packages=['lupyne', 'lupyne.engine'],
    install_requires=['six'],
    extras_require={'server': ['cherrypy>=11', 'clients>=0.2'], 'docs': ['nbsphinx', 'jupyter']},
    python_requires='>=2.7',
    tests_require=['pytest-cov'],
    classifiers=[
        'Development Status :: 6 - Mature',
        'Framework :: CherryPy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Indexing',
    ],
)
